/**
 * \file Event.cpp
 *
 *  \date Dec 20, 2012
 *  \author Alexandre Tuleu
 */

#include "Event.h"

#include <stdexcept>

#include <string>

#include <cstring>

#include <time.h>

#include <errno.h>

namespace rtio {

namespace phidgets {

	Event::Event(Mask iValue)
		: d_mask(iValue){
		int res = pthread_mutex_init(&d_mutex,NULL);
		if(res){
			throw std::runtime_error("Could not create mutex : " + std::string(strerror(res)));
		}

		res = pthread_cond_init(&d_condition,NULL);
		if(res){
			pthread_mutex_destroy(&d_mutex);
			throw std::runtime_error("Could not create condition : " + std::string(strerror(res)));
		}
	}

	Event::~Event(){
		pthread_cond_destroy(&d_condition);
		pthread_mutex_destroy(&d_mutex);
	}

#define WAIT_FOR_MASK(mode,mask,flags) ( (mode == ANY_FLAGS) ? ((d_mask & flags) == 0) : ( (d_mask & flags) != flags) )

	bool Event::Wait(Mask flags,Mask * rMask,Mode m,Timeout t){
		pthread_mutex_lock(&d_mutex);
		if(t < 0){
			if(rMask){
				*rMask = d_mask;
			}
			pthread_mutex_unlock(&d_mutex);
			return true;
		}
		timespec ts;
		if(t) {
			clock_gettime(CLOCK_REALTIME,&ts);
			time_t sec = t/1e3;
			long nsec = (t - (sec * 1e3)) * 1e6;
			ts.tv_sec += sec;
			ts.tv_nsec += nsec;
		}
		bool res = true;
		while(WAIT_FOR_MASK(m,d_mask,flags)){
			if(t){
				int pthread_result = pthread_cond_timedwait(&d_condition,&d_mutex,&ts);
				if( pthread_result == ETIMEDOUT){
					res = false;
					break;
				} else if(pthread_result) {
					pthread_mutex_unlock(&d_mutex);
					throw std::runtime_error("Could not wait on condition : " + std::string(strerror(pthread_result)));
				}
			} else {
				int pthread_result = pthread_cond_wait(&d_condition,&d_mutex);
				if(pthread_result){
					pthread_mutex_unlock(&d_mutex);
					throw std::runtime_error("Could not wait on condition : " + std::string(strerror(pthread_result)));
				}

			}
		}
		if(rMask){
			*rMask = d_mask;
		}
		pthread_mutex_unlock(&d_mutex);
		return res;
	}

	Event::Mask Event::Signal(Event::Mask flags){
		pthread_mutex_lock(&d_mutex);
		Mask res = d_mask;
		d_mask |= flags;
		pthread_cond_signal(&d_condition);
		pthread_mutex_unlock(&d_mutex);
		return res;
	}

	Event::Mask Event::Clear(Mask flags){
		pthread_mutex_lock(&d_mutex);
		Mask res = d_mask;
		d_mask &= ~flags;
		pthread_mutex_unlock(&d_mutex);
		return res;

	}


	void Event::Lock(){
		pthread_mutex_lock(&d_mutex);
	}

	void Event::Unlock(){
		pthread_mutex_unlock(&d_mutex);
	}


} /* namespace phidgets */

} /* namespace rtio */
