/**
 * \file Bridge.cpp
 *
 *  \date Dec 20, 2012
 *  \author Alexandre Tuleu
 */

#include "Bridge.h"

#include "../PhidgetError.h"
#include <sstream>

namespace rtio {

namespace phidgets {

	Bridge::Bridge(int serialNumber, unsigned int timeout)
	: Device<CPhidgetBridgeHandle>(serialNumber,timeout){
		int res = CPhidgetBridge_getInputCount(GetHandle(),&d_maxInput);
		if(res){
			throw PhidgetError("Could not get bridge input count",res);
		}
		res = CPhidgetBridge_getDataRateMin(GetHandle(),&d_maxDataRate);
		if(res){
			throw PhidgetError("Could not get bridge min data rate",res);
		}

		res = CPhidgetBridge_getDataRateMax(GetHandle(),&d_minDataRate);
		if(res){
			throw PhidgetError("Could not get bridge max data rate",res);
		}
		res = CPhidgetBridge_set_OnBridgeData_Handler(GetHandle(),
		                                              &Bridge::DataHandler,
		                                              reinterpret_cast<void*>(this));
		if(res){
			throw PhidgetError("Could not set data handler",res);
		}

		d_data1.assign(d_maxInput,0.0);
		d_data2.assign(d_maxInput,0.0);
		d_dataRdPtr = &d_data1;
		d_dataWrPtr = &d_data2;
	}

	Bridge::~Bridge(){
	}

	int Bridge::MaxInput(){
		return d_maxInput;
	}
	int Bridge::MinDataRate(){
		return d_minDataRate;
	}
	int Bridge::MaxDataRate(){
		return d_maxDataRate;
	}

	CPhidgetBridge_Gain Bridge::GainFromInt(unsigned int gain){
		//get the power of two or 8 if not a power of two <= 128
		unsigned int i;
		for(i = 0 ; i < 8; ++i){
			if(gain & (1U << i) == gain){
				break;
			}
		}


		switch(i){
		case 0 :
			return PHIDGET_BRIDGE_GAIN_1;
		case 3 :
			return PHIDGET_BRIDGE_GAIN_8;
		case 4 :
			return PHIDGET_BRIDGE_GAIN_16;
		case 5 :
			return PHIDGET_BRIDGE_GAIN_32;
		case 6 :
			return PHIDGET_BRIDGE_GAIN_64;
		case 7 :
			return PHIDGET_BRIDGE_GAIN_128;
		default :{
			std::ostringstream os;
			os << "Gain " << gain << " is invalid : it should be in [1 8 16 32 64 128]";
			throw std::out_of_range(os.str());
		}
		}
	}

	void Bridge::SetDataRate(unsigned int millisecond){
		if(millisecond < d_minDataRate || millisecond > d_maxDataRate){
			std::ostringstream os;
			os << "Invalid data rate " << millisecond << " ms : should be"
			   << " in [ " << d_minDataRate << " ms , "<< d_maxDataRate
			   << " ms ]";
			throw std::out_of_range(os.str());
		}
		int res = CPhidgetBridge_setDataRate(GetHandle(),millisecond);
		if(res){
			throw PhidgetError("Could not set data rate",res);
		}
	}


	void Bridge::EnableInput(int input,CPhidgetBridge_Gain gain){
		if(input >= d_maxInput){
			std::ostringstream os;
			os << "Cannot enable input " << input << " : Bridge has only "
			   << d_maxInput << " inputs";
			throw std::out_of_range(os.str());
		}

		if(d_mask & (1 << input)){
			throw std::logic_error("Input is already enabled");
		}

		int res;
		res = CPhidgetBridge_setEnabled(GetHandle(),input,PTRUE);
		if(res){
			throw PhidgetError("Could not enable input",res);
		}
		res = CPhidgetBridge_setGain(GetHandle(),input,gain);
		if(res){
			CPhidgetBridge_setEnabled(GetHandle(),input,PFALSE);
			throw PhidgetError("Could not set input gain",res);
		}

		d_mask |= (1 << input);
	}

	void Bridge::StartFullRefresh(){
		d_event.Clear(d_mask);
	}

	void Bridge::WaitForFullRefresh(){
		d_event.Wait(d_mask,NULL,Event::ALL_FLAGS,Event::INFINITE);
		d_event.Lock();
		std::swap(d_dataWrPtr,d_dataRdPtr);
		d_event.Unlock();

	}

	double Bridge::Data(unsigned int i) const{
		if(i >= d_maxInput){
			std::ostringstream os;
			os << "Cannot access input " << i << " : Bridge has only "
			   << d_maxInput << " inputs";
			throw std::out_of_range(os.str());
		}

		if(d_mask & (1 << i) == 0){
			std::ostringstream os;
			os << "Input " << i << " is not enabled";
			throw std::logic_error(os.str());
		}
		return d_dataRdPtr->at(i);
	}

	int Bridge::DataHandler(CPhidgetBridgeHandle b, void * itselfPtr,int input,double value){
		//get itself object
		Bridge *itself = reinterpret_cast<Bridge*>(itselfPtr);

		//we check if we already writted
		Event::Mask m;
		//this will just inquire, no wait here
		itself->d_event.Wait(0,&m,Event::ANY_FLAGS,Event::NON_BLOCKING);
		if(m & (1 << input) == 0){
			itself->d_dataWrPtr->at(input) = value;
			return 0;
		} //otherwise we could be in the middle of the swap operation of all
		  //bridges that occurs at Bridge::WaitForFullrefresh()
		//now we signal that the input is ready
		itself->d_event.Signal(1 << input);
		return 0;

	}
} /* namespace phidgets */

} /* namespace rtio */
