/**
 * \file MotorQueue.cpp
 *
 *  \date Dec 17, 2012
 *  \author Alexandre Tuleu
 */

#include "MotorQueue.h"
#include "IO.h"


#include <libsbcp/bus/ScheduledWorkflow.h>

namespace rtio {
namespace sbcp {

MotorQueue::MotorQueue(rtio::Manager & manager,
        	           rtio::Queue::ID id,
        	           int priority,
        	           rtio::Queue::Mode mode)
	: rtio::Queue(manager,id,priority,mode){
}

void MotorQueue::SetBus(const std::tr1::shared_ptr< ::sbcp::Bus> & b){
	d_bus = b;
}

MotorQueue::~MotorQueue(){
	Close();
}


void  MotorQueue::AddBoard(const std::string & name,
                           const MotorDriverSection & boardConfig,
                           BrushlessTemplates & templates){
	if(!d_bus){
		throw std::logic_error("Bus should be opened and set !");
	}

	if(boardConfig.BoardID() < 0 || boardConfig.BoardID() > 255){
		std::ostringstream os;
		os << boardConfig.BoardID()
		   << " is not a valid ID (should be in [0 255]";
		throw std::out_of_range(os.str());
	}

	MdvBoardPtr b = d_bus->OpenDevice<MdvBoard>(boardConfig.BoardID());

	if(!b){
		std::ostringstream os;
		os << "Could not find motordirver board with ID "
		    << boardConfig.BoardID();
		throw std::runtime_error(os.str());
	}

	std::tr1::shared_ptr<BrushlessMotorSection>
	m1Template(templates.SubSection(boardConfig.M1Template())),
	m2Template(templates.SubSection(boardConfig.M2Template()));
	if(!m1Template){
		throw std::out_of_range("Could not find '" + boardConfig.M1Template() + "' brushless templates");
	}

	if(!m2Template){
		throw std::out_of_range("Could not find '" + boardConfig.M2Template() + "' brushless templates");
	}

	AddMotor(name + "_" + boardConfig.M1Suffix(),b->Motor1(),*m1Template);
	AddMotor(name + "_" + boardConfig.M2Suffix(),b->Motor2(),*m2Template);

	RegisterInput(name + "_ME1",
	              rtio::sbcp::Input< ::sbcp::Data >::Create(b->Q1().PositionAndStatus()));
	RegisterInput(name + "_ME2",
	              rtio::sbcp::Input< ::sbcp::Data >::Create(b->Q2().PositionAndStatus()));
	RegisterInput(name + "_ME3",
	              rtio::sbcp::Input< ::sbcp::Data >::Create(b->Q3().PositionAndStatus()));

	RegisterInput(name + "_FS1",
		              rtio::sbcp::Input< ::sbcp::Data >::Create(b->Force(0)));
	RegisterInput(name + "_FS2",
		              rtio::sbcp::Input< ::sbcp::Data >::Create(b->Force(1)));
	RegisterInput(name + "_FS3",
		              rtio::sbcp::Input< ::sbcp::Data >::Create(b->Force(2)));

	d_boards.push_back(b);

}

void MotorQueue::AddMotor(const std::string & name,
                          MdvBoard::Motor & motor,
                          const BrushlessMotorSection & config){

	motor.Stiffness().Set(config.Stiffness());
	motor.Damping().Set(config.Damping());
	motor.Preload().Set(config.PreCompression());

	motor.MaxSpeed().Set(config.MaxSpeed());
	motor.MaxAcceleration().Set(config.MaxAcceleration());
	motor.MaxTorque().Set(config.MaxTorque());

	motor.PGain().Set(config.PGain());
	motor.IGain().Set(config.IGain());
	motor.DGain().Set(config.DGain());

	//motor.GoalVelocity()(config.MaxSpeed());

	motor.SmoothPositionUpdate().Set(config.RefreshRate());
	d_controlModes.insert(std::make_pair(&motor,config.GetControlMode()));


	RegisterInput(name,rtio::sbcp::Input< double , int16_t>::Create(motor.PresentPosition()));
	RegisterOutput(name,rtio::sbcp::Output< double, int16_t >::Create(motor.GoalPosition()));

}



void MotorQueue::Close(){
	//resets all the motor to coast mode ....
	d_bus->Lazy();

	for(ListOfMdvBoard::iterator b = d_boards.begin();
	    b != d_boards.end();
	    ++b){
		(*b)->Motor1().MotorControlMode().Set(MdvBoard::Motor::COAST);
		(*b)->Motor2().MotorControlMode().Set(MdvBoard::Motor::COAST);
	}
}


void MotorQueue::FetchUncalibratedBoard(SetOfMotor & uncalibratedBoards){
	uncalibratedBoards.clear();
	for(ListOfMdvBoard::iterator b = d_boards.begin();
	    b != d_boards.end();
	    ++b){
		std::ostringstream os;
		os << (int) (*b)->ID();
		std::string name(os.str());
		if(!(*b)->Motor1().RuntimeCalibrated()){
			uncalibratedBoards.insert(std::make_pair(&(*b)->Motor1(),name +".M1"));
		}
		if(!(*b)->Motor2().RuntimeCalibrated()){
			uncalibratedBoards.insert(std::make_pair(&(*b)->Motor2(),name +".M2"));
		}
	}
}


void MotorQueue::CalibrateMotors(){

	SetOfMotor uncalibrated;
	FetchUncalibratedBoard(uncalibrated);

	if(uncalibrated.empty()){
		PrepareForLoop();
		return;
	}

	for(ListOfMdvBoard::iterator b = d_boards.begin();
	    b != d_boards.end();
	    ++b){
		if((*b)->Motor1().RuntimeCalibrated() && (*b)->Motor2().RuntimeCalibrated()){
			continue;
		}
		(*b)->CalibrateMotors();
	}

	unsigned int count(5);

	while(!uncalibrated.empty()){
		if(++count >= 5){
			std::cout << "waiting for " << uncalibrated.size()
			          << " motor to calibrate : [ ";
			for(SetOfMotor::const_iterator m = uncalibrated.begin();
			    m != uncalibrated.end();
			    ++m){
				if(m != uncalibrated.begin()){
					std::cout << " , ";
				}
				std::cout << m->second;
			}
			count = 0;
		}
		std::cout << std::endl;
		rt_task_sleep(0.2 * 1e9);//sleep 200 ms
		FetchUncalibratedBoard(uncalibrated);
	}

	PrepareForLoop();
}

void MotorQueue::PrepareForLoop(){
	for(ControlModeByMotor::const_iterator cm = d_controlModes.begin();
	    cm != d_controlModes.end();
	    ++cm){
		cm->first->MotorControlMode().Set(cm->second);
	}


	::sbcp::ScheduledWorkflow & w = d_bus->Scheduled();
	for(ListOfMdvBoard::iterator b = d_boards.begin();
	    b != d_boards.end();
	    ++b){
		w.AppendScheduledDevice(*b);
	}
}

void MotorQueue::PreSend(){

}

void MotorQueue::PreRead(){
	::sbcp::ScheduledWorkflow & w = d_bus->Scheduled();
	try{
		w.StartTransfers();
		w.WaitForTransfersCompletion();
	} catch (const ::sbcp::MultipleTransferError & e) {
	}
}

void MotorQueue::PostRead(){}
} /* namespace sbcp */
} /* namespace rx */
