/**
 * \file CdnProcess.cpp
 *
 *  \date Dec 10, 2012
 *  \author Alexandre Tuleu
 */

#include "CdnProcess.h"

#include <codyn/cdn-object.h>

#include <stdexcept>

namespace rx {

	CdnProcess::CdnProcess(const CdnSection & config){
		g_type_init();

		GError * error = NULL;

		d_network = CdnNetworkPtr(cdn_network_new_from_path(config.File().c_str(),&error),
		                          g_object_unref);

		if(!d_network || error){
			std::runtime_error err(std::string("Could not load network file '")
			+ config.File() + std::string("' : ") + error->message);
			g_error_free(error);
			throw err;
		}

		InitOverride(config.Overrides());

		typedef std::tr1::shared_ptr<CdnCompileError> CdnCompileErrorPtr;
		typedef std::tr1::shared_ptr<gchar> GCharPtr;
		//this will be freed when needed
		CdnCompileErrorPtr err= CdnCompileErrorPtr(cdn_compile_error_new(),
		                                     g_object_unref);

		if(!cdn_object_compile(CDN_OBJECT(d_network.get()),NULL,err.get())){
			// this will be freed at throw :)
			GCharPtr msg(cdn_compile_error_get_formatted_string(err.get()),
			             g_free);
			throw std::runtime_error(std::string("Could not compile network : ") + msg.get());
		}
		InitIO();
	}

	CdnProcess::~CdnProcess(){
	}


	void CdnProcess::Step(double ts){
		cdn_network_step(d_network.get(),ts);
	}

	void CdnProcess::InitIO(){
		CdnNode * globals = CDN_NODE(d_network.get());

		GSList * propListPtr = cdn_node_find_variables(globals,
		                                               "has-flag(in)");

		while(propListPtr != NULL){
			CdnVariable * p = CDN_VARIABLE(propListPtr->data);
			this->AddControlInput(cdn_variable_get_name(p),
			                      rx::Input::Ptr(new CdnProcess::Input(p)),
			                      cdn_variable_get_value(p));
			propListPtr = propListPtr->next;
		}
	}

	void CdnProcess::InitOverride(const options::Section & overrides){
		typedef options::Section::ListOfOptions LO;
		const LO & options = overrides.Options();
		CdnVariable * var;
		bool isGlobal;
		for(LO::const_iterator o = options.begin();
		    o != options.end();
		    ++o){
			FindFromSerialized(o->second->Name(),isGlobal,&var);
			if(!var){
				throw std::runtime_error("Could not find variable '"
				                         + o->second->Name() + "'");
			}
			CdnExpression * e = cdn_expression_new(o->second->ValueAsString().c_str());
			cdn_variable_set_expression(var,e);
		}
	}

	rx::Input::Ptr CdnProcess::GetInput(const std::string & name){
		CdnVariable * var;
		bool isGlobal;
		FindFromSerialized(name,isGlobal,&var);
		if(!isGlobal && var == NULL){
			FindFromSerialized(name + "::input",isGlobal,&var);
		}
		if(!var){
			return rx::Input::Ptr();
		}
		return rx::Input::Ptr(new CdnProcess::Input(var));
	}

	rx::Output::Ptr CdnProcess::GetOutput(const std::string & name){
		CdnVariable * var;
		bool isGlobal;
		FindFromSerialized(name,isGlobal,&var);
		if(!isGlobal && var == NULL){
			FindFromSerialized(name + "::output",isGlobal,&var);
		}
		if(!var){
			return rx::Output::Ptr();
		}
		return rx::Output::Ptr(new CdnProcess::Output(var));
	}


	CdnProcess::Input::Input(CdnVariable * var)
		: d_variable(var){

	}

	CdnProcess::Input::~Input(){
	}

	void CdnProcess::Input::SendToProcess(double value){
		cdn_variable_set_value(d_variable,value);
	}

	CdnProcess::Output::Output(CdnVariable * var)
		: d_variable(var){
	}

	CdnProcess::Output::~Output(){}

	double CdnProcess::Output::ReadFromProcess(){
		return cdn_variable_get_value(d_variable);
	}


	void CdnProcess::FindFromSerialized(const std::string & name,
	                                    bool & isNameGlobal,
	                                    CdnVariable ** var){
		std::string::size_type pos = name.find("::");
		std::string objname,propname;
		CdnObject * object;
		if(pos == std::string::npos){
			isNameGlobal = true;
			object = CDN_OBJECT(d_network.get());
			propname = name;
		} else {
			objname = name.substr(0,pos);
			propname = name.substr(pos+2);
			object = cdn_node_get_child(CDN_NODE(d_network.get()),objname.c_str());
		}

		if(!object){
			(*var) = NULL;
		} else {
			(*var) = cdn_object_get_variable(object,propname.c_str());
		}
	}

} /* namespace rx */
