/**
 * \file Interface.cpp
 *
 *  \date Dec 10, 2012
 *  \author Alexandre Tuleu
 */

#include "Interface.h"
#include <cstring>

#include <robo-xeno/NativeError.h>

namespace rx {

	Interface::Interface()
		: cli::Interface(1,2,false)
		, d_process()
		, d_runner(NULL)
		, d_managerDisplay(AddWidget<ManagerDisplayWidget>("RT I/O",0,0,1,1))
		, d_variableTable(AddWidget<VariableTable<double> >("Process Controll Variable",0,1,1,1)){
		def_prog_mode();
		endwin();
		Init();
		d_task = RtTaskPtr(new RT_TASK(),rt_task_delete);
		int res = rt_task_create(d_task.get(),NULL,0,0,0);
		if(res){
			throw NativeError("Could not create task",-res);
		}
	}

	Interface::~Interface(){
	}

	void Interface::Init(){
		if(cli::Interface::HasColor()){
			if(ERR == init_pair(BlackOnWhite,COLOR_BLACK,COLOR_WHITE)){
				endwin();
				throw std::runtime_error("Could not init color pair BlackOnWhite");
			}
			if(ERR == init_pair(WhiteOnBlue,COLOR_BLACK,COLOR_BLUE)){
				endwin();
				throw std::runtime_error("Could not init color pair BlackOnWhite");
			}
		}

		AddCommand(cli::Closure1<rx::Interface,std::string>::Create("save",
		                                                            *this,
		                                                            &rx::Interface::Save));

		AddCommand(cli::Closure0<rx::Interface>::Create("unselect",
		                                                *this,
		                                                &rx::Interface::Unselect));

		AddCommand(cli::Closure0<rx::Interface>::Create("start",
		                                                *this,
		                                                &rx::Interface::Start));

		AddCommand(cli::Closure0<rx::Interface>::Create("pause",
		                                                *this,
		                                                &rx::Interface::Pause));

		AddCommand(cli::Closure2<rx::Interface,
		                         std::string,
		                         double>::Create("select",
		                                         *this,
		                                         &rx::Interface::Select));
	}

	void Interface::SetProcess(Process::Ptr process){
		if(d_process){
			return;
		}
		d_process = process;

		Process::ControlInputByName & controlInputs = d_process->ControlInputs();

		for(Process::ControlInputByName::iterator ci = controlInputs.begin();
		    ci != controlInputs.end();
		    ++ci){
			Interface::Variable::Ptr v(new Variable(ci->first,
			                                        ci->second,
			                                        *d_variableTable,
			                                        *d_process));
			this->AddVariable(std::tr1::static_pointer_cast<cli::VariableBase,Interface::Variable>(v));
			d_variableTable->AddVariable(std::tr1::static_pointer_cast<cli::Variable<double>,rx::Interface::Variable >(v));
			d_variables[ci->first] = v;
		}
		d_variableTable->Refresh();
	}


	void Interface::SetManager(rtio::Manager & manager){
		d_managerDisplay->InitContentWith(manager);
	}

	void Interface::SetRunner(ProcessRunner & runner){
		if(d_runner){
			return;
		}
		d_runner = &runner;
	}

	void Interface::Save(const std::string & ){
		std::cerr <<"Unimplemented yet ! " << std::endl;
	}

	void Interface::Select(const std::string & propName,
	                       const double & increment){
		VariableByName::iterator fi = d_variables.find(propName);
		if(fi == d_variables.end()){
			std::cerr << "Could not find variable '" << propName << "'"
			          << std::endl;
			return;
		}
		if(increment == 0){
			std::cerr << "You may want to have an increment > 0 :P" << std::endl;
			return;
		}

		this->AddBinding(IncrementKey,
		                 cli::KeyCallback::Ptr(new IncrementCallback(fi->second,
		                                                             increment)));
		this->AddBinding(DecrementKey,
		                 cli::KeyCallback::Ptr(new IncrementCallback(fi->second,
		                                                             -increment)));
	}


	void Interface::Unselect(){
		this->Unbind(IncrementKey);
		this->Unbind(DecrementKey);
	}


	Interface::Variable::Variable(const std::string & name,
	                              const Process::ControlInput & ci,
	                              VariableTable<double> & vt,
	                              rx::Process & process)
		: cli::Variable<double>(name,ci.value)
		, d_ci(ci)
		, d_vt(vt)
		, d_process(process){
	}

	Interface::Variable::~Variable(){
	}

	void Interface::Variable::PostSetValueHook(){
		d_ci.value = Value();
		d_process.BufferedWrite(d_ci);
		SwitchToSecondary();
		d_vt.UpdateVariable(this);
		d_vt.Refresh();
	}


	Interface::IncrementCallback::IncrementCallback(Variable::Ptr v,
	                                                double increment)
		: d_variable(v)
		, d_increment(increment){
	}

	Interface::IncrementCallback::~IncrementCallback(){}

	int Interface::IncrementCallback::Call(int){
		d_variable->SetValue(d_variable->Value() + d_increment);
		return 0;
	}


	void Interface::Start(){
		if(!d_runner){
			std::cerr << "Process is not initialized." << std::endl;
			return;
		}
		d_runner->SetPause(false);
		SwitchToSecondary();
	}

	void Interface::Pause(){
		if(!d_runner){
			std::cerr << "Process is not initialized." << std::endl;
			return;
		}
		d_runner->SetPause(true);
		SwitchToSecondary();
	}

	void Interface::StartTask(){
		if(!d_runner){
			throw std::runtime_error("Runner is not initialized, I will not start !!");
		}
		reset_prog_mode();
		rt_task_start(d_task.get(),&Interface::TaskBody,reinterpret_cast<void*>(this));
	}

	void Interface::TaskBody(void * itself){
		reinterpret_cast<rx::Interface*>(itself)->TaskLoop();
	}

	void Interface::TaskLoop(){
		while(true){
			Process();
			if(UserAskedToQuit()){
				break;
			}
			d_managerDisplay->RefreshIfNeeded();
		}
		if(d_runner){
			d_runner->StopTask();
		}

	}

	void Interface::SwitchToSecondary(){
		rt_task_set_mode(T_PRIMARY,0,NULL);
	}
} /* namespace rx */
