/**
 * \file ManagerDisplayWidget.cpp
 *
 *  \date Dec 10, 2012
 *  \author Alexandre Tuleu
 */

#include "ManagerDisplayWidget.h"
#include <robo-xeno/NativeError.h>

#include <sstream>
#include <iomanip>

#include "Interface.h"

namespace rx {
	const std::string ManagerDisplayWidget::Unplugged(" ??? ");
	const std::string ManagerDisplayWidget::Plugged(" --> ");

	ManagerDisplayWidget::ManagerDisplayWidget(Def & def)
		: cli::Widget(def)
		, ColSize((Width() - ProcessBlockWidth - 2 * (Unplugged.size() + 1))/2)
		, RowSize(Height() / 2)
		, d_mutex(new RT_MUTEX())
		, d_refresh(false){
		int res = rt_mutex_create(d_mutex.get(),NULL);
		if(res){
			endwin();
			throw NativeError("Could not create mutex",-res);
		}
	}

	ManagerDisplayWidget::~ManagerDisplayWidget(){
	}

	void ManagerDisplayWidget::InitContentWith(rtio::Manager & manager){
		LogQueue::Ptr l = manager.AddLogQueue<LogQueue>();
		((ManagerDisplayWidget::LogQueue&)(*l)).SetManagerDisplay(this);
		DrawCenter();
	}
	void ManagerDisplayWidget::DrawBox(int x, int y, int width, int height){
		mvwvline(Window(),y+ 1,x,ACS_VLINE,height -2);
		mvwvline(Window(),y + 1,x + width - 1,ACS_VLINE,height- 2);
		mvwhline(Window(),y,x + 1,ACS_HLINE,width-2);
		mvwhline(Window(),y + height -1,x + 1 ,ACS_HLINE,width-2);
		mvwaddch(Window(),y,x,'+');
		mvwaddch(Window(),y,x + width - 1,'+');
		mvwaddch(Window(),y + height - 1,x + width - 1,'+');
		mvwaddch(Window(),y + height - 1,x ,'+');
	}

	void ManagerDisplayWidget::VPrint(int x, int y, const std::string & name){
		for(unsigned int i = 0; i < name.size(); ++i){
			mvwaddch(Window(),(int)(y + (int)i),(int)x,(chtype)(name.at(i)));
		}
	}
	void ManagerDisplayWidget::HPrint(int x, int y, const std::string & name){
			for(unsigned int i = 0; i < name.size(); ++i){
				mvwaddch(Window(),(int)(y),(int)(x + (int)i),(chtype)(name.at(i)));
			}
		}

	void ManagerDisplayWidget::DrawCenter(){

		DrawBox(1 + ColSize + Unplugged.size(),0,ProcessBlockWidth,Height());

		std::string name("PROCESS");

		VPrint(1 + ColSize + Unplugged.size() + 2,(Height()-2-name.size())/2,name);

		std::vector<std::string> prefixes;
		prefixes.push_back("left_fore");
		prefixes.push_back("right_fore");
		prefixes.push_back("left_hind");
		prefixes.push_back("right_hind");

	}

	ManagerDisplayWidget::NameAndRow::NameAndRow(const std::string & name_,int row_,unsigned int size)
		: name(name_)
		, row(row_){
		CutNameIfNeeded(size,name);
	}

	void ManagerDisplayWidget::CutNameIfNeeded(unsigned int size,
	                                           std::string & str){
		if(str.size() > size){
			str = str.substr(0,size - 3) + "...";
		}
	}


	void ManagerDisplayWidget::PrintInput(const NameAndRow & nr, double value){
		std::ostringstream os;
		os << std::setw((int)ColSize) << std::left << nr.name << Plugged;
		HPrint(1,nr.row,os.str());
		std::ostringstream osVal;
		osVal << std::setprecision((int)(ColSize + Plugged.size()))
		      << std::setw((int)(ColSize+ Plugged.size()))
		      << std::left << value;
		HPrint(1,nr.row + 1,osVal.str());

	}

	void ManagerDisplayWidget::PrintOutput(const NameAndRow & nr, double value){
		std::ostringstream os;
		os << Plugged << std::setw((int)ColSize) << std::right << nr.name;
		HPrint((int)(1 + ColSize + ProcessBlockWidth + Plugged.size()) ,nr.row,os.str());
		std::ostringstream osVal;
		osVal << std::setprecision((int)(ColSize + Plugged.size()))
		      << std::setw((int)(ColSize + Plugged.size()))
		      << std::right << value;
		HPrint((int)(1 + ColSize + ProcessBlockWidth + Plugged.size()) ,nr.row + 1,osVal.str());

	}

	void ManagerDisplayWidget::PostponedRefresh(){
		rt_mutex_acquire(d_mutex.get(),TM_INFINITE) ;
		d_refresh = true;
		rt_mutex_release(d_mutex.get());
	}
	void ManagerDisplayWidget::RefreshIfNeeded(){
		rt_mutex_acquire(d_mutex.get(),TM_INFINITE) ;
		bool actuallyRefresh = d_refresh;
		rt_mutex_release(d_mutex.get());
		rt_task_set_mode(T_PRIMARY,0,NULL);
		if(actuallyRefresh){
			wrefresh(Window());
		}
	}

	ManagerDisplayWidget::LogQueue::LogQueue(ID id)
		: rtio::LogQueue(id)
		, d_display(NULL){

	}
	ManagerDisplayWidget::LogQueue::~LogQueue(){
	}
	void ManagerDisplayWidget::LogQueue::SetManagerDisplay(ManagerDisplayWidget * md){
		d_display = md;
	}
	void ManagerDisplayWidget::LogQueue::Init(){
	}
	void ManagerDisplayWidget::LogQueue::Process(){

		unsigned int maxRow = (d_display->Height()-1)/ 2;
		unsigned int displayedInputs = std::min(maxRow,(unsigned int)InputNames().size());
		for(unsigned int i = 0; i < displayedInputs; ++i){
			NameAndRow nr(InputNames()[i],(int)(2 * i) + 1,d_display->ColSize);
			d_display->PrintInput(nr,InputValues()[i]);
		}

		unsigned int displayedOutputs = std::min(maxRow,(unsigned int)OutputNames().size());
		for(unsigned int i = 0; i < displayedOutputs; ++i){
			NameAndRow nr(OutputNames()[i],(int)(2 * i) + 1,d_display->ColSize);
			d_display->PrintOutput(nr,OutputValues()[i]);
		}


		d_display->RefreshIfNeeded();
	}

} /* namespace rx */
