/**
 * \file VariableTable.h
 *
 *  \date Dec 10, 2012
 *  \author Alexandre Tuleu
 */

#ifndef ROBO_XENO_VARIABLETABLE_H_
#define ROBO_XENO_VARIABLETABLE_H_

#include <map>
#include <set>
#include <iostream>


#include <biorob-cli/Widget.h>
#include <biorob-cli/Variable.h>
#include <biorob-cli/Interface.h>

#include <stdexcept>
#include <iomanip>

#include "color.h"
/**
 *
 */
namespace rx {
/**
 *
 */
template <typename T>
class VariableTable : public cli::Widget {
public:
	typedef std::tr1::shared_ptr<VariableTable>         Ptr;
	typedef cli::Variable<T>                            VariableType;
	typedef typename std::tr1::shared_ptr<VariableType> VariableTypePtr;
	VariableTable(Def & def);
	virtual ~VariableTable();

	void AddVariable(VariableTypePtr v);

	void UpdateVariable(VariableType const * v);

	void Refresh();

private :
	typedef std::set<VariableTypePtr>      SetOfVariable;
	typedef std::map<VariableType const *,int >   RowByVariable;
	typedef typename RowByVariable::const_iterator RowByVariableConstIterator;

	void PrintVariable(RowByVariableConstIterator i);
	RowByVariable d_rows;
	SetOfVariable d_variables;
	const unsigned int NumberSize;
	const unsigned int MaxNameSize;
	const bool   UseColor;
};

template <typename T>
inline VariableTable<T>::VariableTable( Def & def)
	: cli::Widget(def)
	, NumberSize(std::min((Width() - 5)/2,(std::size_t)15U))
	, MaxNameSize(Width() - 5 - NumberSize)
	, UseColor(cli::Interface::HasColor()){
	if(Width() - 5 < 10){
		throw std::runtime_error("Unsufficient size for widget.");
	}

}

template <typename T>
inline VariableTable<T>::~VariableTable(){}


template <typename T>
inline void VariableTable<T>::AddVariable(VariableTypePtr v){
	if(d_rows.size() >= Height()){
		std::cerr << "Could not display variable '" << v->Name() << std::endl;
		return;
	}

	if( d_variables.find(v) != d_variables.end()){
		return;
	}
	d_variables.insert(v);
	PrintVariable((d_rows.insert(std::make_pair(v.get(),d_rows.size()))).first);

}


template <typename T>
inline void VariableTable<T>::UpdateVariable(VariableType const * v){
	RowByVariableConstIterator fi = d_rows.find(v);

	if(fi == d_rows.end()){
		return;
	}

	PrintVariable(fi);
}

template <typename T>
inline void VariableTable<T>::Refresh(){
	wrefresh(Window());
}

template <typename T>
inline void VariableTable<T>::PrintVariable(RowByVariableConstIterator i){
	std::ostringstream os;
	os << std::left << std::setw((int)MaxNameSize);
	if(i->first->Name().size() > MaxNameSize){
		os << (i->first->Name().substr(0,MaxNameSize - 3) + "...");
	} else {
		os << i->first->Name();
	}
	os << " = "
	   << std::setprecision((int)(NumberSize-2))
	   << std::setw((int)NumberSize) << std::left << i->first->Value();

	wmove(Window(),i->second,1);
	wclrtoeol(Window());
	wmove(Window(),i->second,1);

	if(UseColor){
		int color = (i->second % 2) ? COLOR_PAIR(rx::BlackOnWhite)
		                            : COLOR_PAIR(cli::Interface::WhiteOnBlack);
		wattrset(Window(),color);
	}

	waddstr(Window(),os.str().c_str());

}
} //namespace rx
#endif // ROBO_XENO_VARIABLETABLE_H_
