/**
 * \file ProcessRunner.cpp
 *
 *  \date Dec 10, 2012
 *  \author Alexandre Tuleu
 */

#include <robo-xeno/NativeError.h>
#include "ProcessRunner.h"
#include <native/task.h>
#include <iostream>

#include <robo-xeno/interface/Interface.h>
#include <cstring>

#include <algorithm>

#include <robo-xeno/rtio/FileLogQueue.h>

#include <biorob-cpp/log/Logger.h>

namespace rx {

	ProcessRunner::ProcessRunner(const MainConfigSection & config)
		: d_config(config)
		, d_task(new RT_TASK())
		, d_period(d_config.Timestep())
		, d_onPause(true){
		int res = rt_task_create(d_task.get(),"robo-xeno-main",0,config.Priority(),T_JOINABLE);
		if(res){
			throw NativeError("Got error during task creation : ",-res);
		}
		d_mutex = NativeHolder<RT_MUTEX>(new RT_MUTEX());
		res = rt_mutex_create(d_mutex.get(),NULL);
		if(res){
			throw NativeError("Could not create mutex",-res);
		}
	}

	ProcessRunner::~ProcessRunner(){
	}

	void ProcessRunner::Task(void * itselfPtr){
		ProcessRunner * itself = reinterpret_cast<ProcessRunner*>(itselfPtr);
		itself->MainInit();
		itself->MainLoop();
	}

	void ProcessRunner::MapInput(){
		//auto mapping the input
		const rtio::Queue::InputByName &  rtioInputs  = d_manager.Inputs();
		//we would like to insure unicity of mapping : only one rtio::Input
		//is mapped to a rx::Input. so we use a nice map.
		typedef std::map<rx::Input::Ptr,rtio::Input::Ptr> InputMapping;
		InputMapping  inputMapping;
		for(rtio::Queue::InputByName::const_iterator ni = rtioInputs.begin();
		    ni != rtioInputs.end();
		    ++ni){
			rx::Input::Ptr pi = d_process->GetInput(ni->first);
			if(pi){
				inputMapping[pi] = ni->second;
				std::cout << "Auto-mapped input '" << ni->first << "'" << std::endl;
			}

		}

		MappingSection::ProcessIOByRTIO inMapping;
		d_config.Mapping().In().Map(inMapping);

		for(MappingSection::ProcessIOByRTIO::const_iterator rt_p = inMapping.begin();
		    rt_p != inMapping.end();
		    ++rt_p){
			std::ostringstream os;
			os << "Error during input mapping rt:'" << rt_p->first
			   << "' <-> rx:'" << rt_p->second << "' : " << std::endl;
			bool throwExcept(false);
			rtio::Input::Ptr rt_i = d_manager.GetInput(rt_p->first);


			if(!rt_i){
				throwExcept = true;
				os << "* Could not find rtio:'" << rt_p->first << "'";
			}

			rx::Input::Ptr p_i = d_process->GetInput(rt_p->second);
			if(!p_i){
				if(throwExcept){
					os << std::endl;
				}
				throwExcept = true;
				os << "* Could not find rx:'" << rt_p->second << "'";
			}

			if(throwExcept){
				throw std::runtime_error(os.str());
			}

			inputMapping[p_i] = rt_i;
			std::cout << "Mapped input rtio:'" <<rt_p->first
			          << "' to rx:'" << rt_p->second <<"'" << std::endl;
		}

		//actually do the mapping;
		for(InputMapping::const_iterator im = inputMapping.begin();
		    im != inputMapping.end();
		    ++im){
			im->second->SetProcessInput(im->first);
		}
	}

	void ProcessRunner::MapOutput(){

		//automapping the output
		const rtio::Queue::OutputByName & rtioOutputs = d_manager.Outputs();

		for(rtio::Queue::OutputByName::const_iterator no = rtioOutputs.begin();
		    no != rtioOutputs.end();
		    ++no){
			rx::Output::Ptr po = d_process->GetOutput(no->first);
			if(po){
				no->second->SetProcessOutput(po);
				std::cout << "Auto-mapped output '" << no->first << "'"
				          << std::endl;
			}
		}


		MappingSection::ProcessIOByRTIO outMapping;
		d_config.Mapping().Out().Map(outMapping);

		for(MappingSection::ProcessIOByRTIO::const_iterator rt_p = outMapping.begin();
		    rt_p != outMapping.end();
		    ++rt_p){
			std::ostringstream os;
			os << "Error during output mapping rt:'"  << rt_p->first
			   << "' <-> rx:'" << rt_p->second << " : "  <<std::endl;
			bool throwExcept(false);

			rtio::Output::Ptr rt_o = d_manager.GetOutput(rt_p->first);

			if(!rt_o){
				throwExcept = true;
				os << "* Coulf not find rt:'" << rt_p->first << "'";
			}

			rx::Output::Ptr p_o = d_process->GetOutput(rt_p->second);

			if(!p_o){
				if(throwExcept){
					os << std::endl;
				} else {
					throwExcept = true;
				}
				os << "* Could not find rx:'" << rt_p->second << "'";
			}

			if(throwExcept){
				throw std::runtime_error(os.str());
			}

			rt_o->SetProcessOutput(p_o);

			std::cout << "Map output rx:'" <<rt_p->second
			          << "' to rtio:'" << rt_p->first <<"'" << std::endl;
		}


		//small optimization step : for all outputs, we store in a vector
		//the mapped one, to avoid call to the one unmapped.

	}

	void ProcessRunner::MainInit(){
		if(!d_process){
			throw std::logic_error("Process is not initialized.");
		}

		d_moduleLoader.LoadModuleFromConfig(d_config.Modules());

		d_moduleLoader.StartQueuesOfModules(d_manager);

		//here we are sure all plugin are loaded and IO made
		MapInput();
		MapOutput();

		std::ostringstream os;
		os << "Error during configuration : " << std::endl;
		bool throwExcept(false);

		if(! d_config.IOLog().empty() ){
			rtio::LogQueue::Ptr q = d_manager.AddLogQueue<rtio::FileLogQueue>();

			((rtio::FileLogQueue*)(q.get()))->SetFileName(d_config.IOLog());
		}

		if(! d_config.QueueSynchronizationLog().empty()) {
			if(throwExcept){
				os << std::endl;
			} else {
				throwExcept = true;
			}
			os << "* queue-synchro-log is unimplemented yet !";
		}

		if(throwExcept){
			throw std::runtime_error(os.str());
		}

		d_interface->SetRunner(*this);
		d_interface->SetProcess(d_process);
		d_interface->SetManager(d_manager);

		d_manager.InitializeQueues();

		d_interface->StartTask();
	}

	void ProcessRunner::MainLoop(){
		unsigned long overruns(42);
		unsigned int nbIteration(std::max(1U,d_config.IterationPerStep()));
		int logEveryIteration(d_config.CliFeedbackIteration());
		unsigned int iterCount;
		unsigned int paused = 0;
		bool logThisTime(false);
		double time(0);
		//we wait a first time, because init of interface, log queues and so
		//could have taken lot of time, but we don't care ...
		int res = rt_task_wait_period(&overruns);
		if(res != 0 && res != -ETIMEDOUT){
			endwin();
			throw NativeError("Could not rt_task_wait_period : " ,-res);
		}
		//now we want hard realtime, no memory allocation, no syscall ...
		while(true){

			res = rt_task_wait_period(&overruns);
			if(res != 0 && res != -ETIMEDOUT){
				endwin();
				throw NativeError("Could not rt_task_wait_period : " ,-res);
			}

			iterCount += overruns + 1;
			time += (overruns + 1) * d_period;

			const rtio::Manager::ListOfQueue & idles = d_manager.IdleQueues();
			for(rtio::Manager::ListOfQueue::const_iterator q = idles.begin();
			    q != idles.end();
			    ++q){
				const rtio::Queue::InputByName & inputs = (*q)->Inputs();
				for(rtio::Queue::InputByName::const_iterator i = inputs.begin();
				    i != inputs.end();
				    ++i){
					if(i->second->IsMapped()){
						i->second->SendToProcess();
					}
				}
			}




			for(unsigned int i = 0; i < (paused * nbIteration * (overruns + 1)); ++i){
				d_process->Step(d_period / ((double) nbIteration));
			}


			for(rtio::Manager::ListOfQueue::const_iterator q = idles.begin();
			    q != idles.end();
			    ++q){
				const rtio::Queue::OutputByName & outputs = (*q)->Outputs();
				for(rtio::Queue::OutputByName::const_iterator o = outputs.begin();
				    o != outputs.end();
				    ++o){
					if(o->second->IsMapped()){
						o->second->ReadFromProcess();
					}
				}
			}

			if(logEveryIteration > 0 && iterCount >= (unsigned int)logEveryIteration){
				logThisTime = true;
				iterCount -= (unsigned int)logEveryIteration;
				d_manager.BufferData(time);
			} else {
				logThisTime = false;
			}


			d_manager.StartIdleQueues();

			if(logThisTime){
				d_manager.StartIdleLogQueues();
			}

			//acquire the control mutex to
			if(!rt_mutex_acquire(d_mutex.get(),TM_NONBLOCK)){
				paused = d_onPause ? 0 : 1;
				rt_mutex_release(d_mutex.get());
			}

			if(overruns){
				std::cerr << "Got " << overruns << " overrun";
				if(overruns > 1){
					std::cerr << "s";
				}
				std::cerr << std::endl;
			}

			//flush the pending write during sync
			d_process->Flush();
		}
	}

	void ProcessRunner::StartTask(){
		int res =rt_task_set_periodic(d_task.get(),TM_NOW,d_period * 1e9);
		if(res){
			throw NativeError("Could not set periodicity" ,-res);
		}
		//reinterpret_cast will be reinterpreted again
		rt_task_start(d_task.get(),&ProcessRunner::Task,reinterpret_cast<void*>(this));
	}

	void ProcessRunner::StopTask(){
		//stop the interface ....
		d_interface = InterfacePtr();
		//stop the task
		d_task = NativeHolder<RT_TASK>();
	}

	void ProcessRunner::SetProcess(Process::Ptr process){
		d_process = process;
	}

	void ProcessRunner::SetInterface(InterfacePtr i){
		d_interface = i;
	}

	void ProcessRunner::WaitForCompletion(){
		rt_task_join(d_task.get());
	}

	void ProcessRunner::SetPause(bool on){
		if(rt_mutex_acquire(d_mutex.get(),TM_INFINITE)){
			return ;
		}
		d_onPause = on;
		rt_mutex_release(d_mutex.get());
	}
} /* namespace rx */
