/*
 * FileLogQueue.cpp
 *
 *  Created on: Dec 20, 2012
 *      Author: tuleu
 */

#include "FileLogQueue.h"
#include <stdexcept>
#include <cstring>

#include <sstream>

namespace rtio {

const DataIO::FormatVersion DataIO::CURRENT_VERSION = 0x01;

DataIO::DataIO(const std::string & filename)
	: d_file(filename.c_str(),std::ios_base::in | std::ios_base::out | std::ios_base::binary)
	, d_inputSize(0)
	, d_outputSize(0){

	if(!d_file.is_open()){
		throw std::runtime_error("Could not open file '" + filename + "' : " + strerror(errno));
	}
}

DataIO::~DataIO(){}

void DataIO::WriteHeader(const LogQueue::ListOfName & inputName,
                         const LogQueue::ListOfName & outputName){
	d_file.write(reinterpret_cast<const char*>(&CURRENT_VERSION),sizeof(FormatVersion));
	PrintListOfString(inputName);
	PrintListOfString(outputName);
	d_inputSize = inputName.size();
	d_outputSize = outputName.size();
}

#define THROW_IF_NOT_GOOD(str) do{\
	if(!d_file.good()){ \
		throw std::runtime_error(str);\
	} \
}while(0)

void DataIO::GetHeader(LogQueue::ListOfName & inputName,
                       LogQueue::ListOfName & outputName){
	FormatVersion fileVersion;
	d_file.read(reinterpret_cast<char*>(&fileVersion),sizeof(FormatVersion));
	THROW_IF_NOT_GOOD("Could not read version");
	if(fileVersion != CURRENT_VERSION){
		std::ostringstream os;
		os << "Unknown file version " << fileVersion << " only " << CURRENT_VERSION << " is supported";
		throw std::runtime_error(os.str());
	}
	ReadListOfString(inputName);
	ReadListOfString(outputName);
	d_inputSize = inputName.size();
	d_outputSize = outputName.size();
}


void DataIO::PrintListOfString(const LogQueue::ListOfName & list){
	FieldLength l(list.size());
	d_file.write(reinterpret_cast<const char *>(&l),sizeof(FieldLength));
	for(unsigned int i = 0; i < l;++i){
		FieldLength sLength = list[i].size();
		d_file.write(reinterpret_cast<const char *>(&sLength),sizeof(FieldLength));
		d_file.write(list[i].c_str(),sLength);
	}
}

void DataIO::ReadListOfString(LogQueue::ListOfName & list){
	FieldLength l;
	d_file.read(reinterpret_cast<char *>(&l),sizeof(FieldLength));
	THROW_IF_NOT_GOOD("Could not read list of string length");
	list.clear();
	list.reserve(l);
	for(unsigned int i = 0; i < l; ++i){
		FieldLength sLength;
		d_file.read(reinterpret_cast<char *>(&sLength),sizeof(FieldLength));
		THROW_IF_NOT_GOOD("Could not read length of string");
		char * value = new char[sLength + 1];
		d_file.read(value,sLength);
		THROW_IF_NOT_GOOD("Could not read string");
		value[sLength] = 0;
		list.push_back(value);
		delete [] value;
	}
}

void DataIO::WriteRow(double time,
                      const DataVector & inputs,
                      const DataVector & outputs){
	bool throwExcept(false);
	std::ostringstream os;
	if(inputs.size() != d_inputSize){
		throwExcept = true;
		os << "* Uncorrect input size : expected " << d_inputSize << " but got "
		   << inputs.size();
	}

	if(outputs.size() != d_outputSize){
		if(throwExcept){
			os << std::endl;
		} else {
			throwExcept = true;
		}
		os << "* Uncorrect output size : expected " << d_outputSize << " but got "
		   << outputs.size();
	}

	if(throwExcept){
		throw std::out_of_range(os.str());
	}

	d_file.write(reinterpret_cast<const char *>(&time),sizeof(double));
	d_file.write(reinterpret_cast<const char *>(inputs.data()),d_inputSize * sizeof(double));
	d_file.write(reinterpret_cast<const char *>(outputs.data()),d_outputSize * sizeof(double));

}

bool DataIO::GetRow(double & time,
                    DataVector & inputs,
                    DataVector & outputs){

	d_file.read(reinterpret_cast<char*>(&time),sizeof(double));
	THROW_IF_NOT_GOOD("Could not read time");
	inputs.clear();
	inputs.assign(d_inputSize,0.0);
	d_file.read(reinterpret_cast<char*>(inputs.data()),d_inputSize * sizeof(double));
	THROW_IF_NOT_GOOD("Could not read input");


	outputs.clear();
	outputs.assign(d_outputSize,0.0);
	d_file.read(reinterpret_cast<char*>(outputs.data()),d_outputSize * sizeof(double));
	return d_file.good();
}


FileLogQueue::FileLogQueue(ID id)
	: LogQueue(id)
	, d_file(){
}

FileLogQueue::~FileLogQueue(){
}

void FileLogQueue::Init(){
	if(!d_file){
		throw std::logic_error("You need to set the filename before");
	}

	d_file->WriteHeader(InputNames(),OutputNames());

}
void FileLogQueue::Process(){
	d_file->WriteRow(Time(),InputValues(),OutputValues());
}

void FileLogQueue::SetFileName(const std::string & filename){
	d_file = std::tr1::shared_ptr<DataIO>(new DataIO(filename));
}

} /* namespace rtio */
