/**
 * \file IOHolder.h
 *
 *  \date Dec 13, 2012
 *  \author Alexandre Tuleu
 */

#ifndef ROBO_XENO_IOHOLDER_H_
#define ROBO_XENO_IOHOLDER_H_

#include "IO.h"

#include <string>
#include <vector>
#include <map>

namespace rtio {
	typedef std::vector<double>        DataVector;

template <typename T>
class IOHolder {
public :
	typedef std::tr1::shared_ptr<T>    TPtr;
	typedef std::map<std::string,TPtr> TByName;
	typedef std::vector<std::string>   NameByIndex;


	IOHolder();
	~IOHolder();


	bool HasIO(const std::string & name) const{
		return d_tByName.find(name) != d_tByName.end();
	}

	TPtr IO(const std::string & name) const{
		typename TByName::const_iterator fi = d_tByName.find(name);
		return fi == d_tByName.end() ? TPtr() : fi->second;
	}

	bool Register(const std::string & name, const TPtr & tptr);

	const TByName & IOByName() const{
		return d_tByName;
	}

	const DataVector & IOData() const{
		return d_data;
	}

	const NameByIndex IONames() const{
		return d_names;
	}

	std::vector<double>::size_type Size() const {
		return d_data.size();
	}

private :
	typedef std::vector<TPtr>          TByIndex;

	TByName     d_tByName;
	NameByIndex d_names;
	DataVector  d_data;
	TByIndex    d_ts;

};


template <typename T> inline
IOHolder<T>::IOHolder(){}

template <typename T> inline
IOHolder<T>::~IOHolder(){}



template <typename T> inline
bool IOHolder<T>::Register(const std::string & name, const TPtr & tptr){
	if(HasIO(name)){
		return false;
	}
	//check if reallocation will happen
	if(d_data.capacity() == d_data.size()){
		//force reallocation now
		d_data.reserve(d_data.capacity() + 1);
		for(typename TByIndex::const_iterator t_it = d_ts.begin();
		    t_it != d_ts.end();
		    ++ t_it){
			unsigned int i = t_it-d_ts.begin();
			//here we just correct the pointer, the reallocation took care
			//to copy the value, so we have to set false (anyway the memory
			//is not reachable anymore)
			(*t_it)->SetValuePtr(&(d_data[i]),false);
		}
	}

	d_data.push_back(0.0);//no reallocation here :)
	d_names.push_back(name);
	d_ts.push_back(tptr);
	d_tByName.insert(std::make_pair(name,tptr));
	tptr->SetValuePtr(&(d_data.back()),true);
	return true;
}










} //namespace rtio
#endif // ROBO_XENO_IOHOLDER_H_
