/**
 * \file RTIOManager.cpp
 *
 *  \date Dec 4, 2012
 *  \author Alexandre Tuleu
 */

#include "Manager.h"
#include <climits>
#include <robo-xeno/NativeError.h>
#include <stdexcept>
#include <sstream>
#include <cstring>


using namespace rtio;

const char * Manager::IOEventName = "rx-manager-io-sync";
const char * Manager::LogEventName = "rx-manager-log-sync";

const Queue::ID Manager::MaxNumberOfQueue = sizeof(unsigned long) * CHAR_BIT;

Manager::Manager()
	: d_outputs()
	, d_inputs()
	, d_idleQueues()
	, d_queues()
	, d_ioEvent()
	, d_allEventMask()
	, d_lastEventMask(){
	d_queues.reserve(MaxNumberOfQueue);
}


Manager::~Manager() {
	d_idleQueues.clear();
	//we stop all the queue, because they will become unsynchronized, so we
	// do not want that
	for(QueueById::iterator q = d_queues.begin();
	    q != d_queues.end();
	    ++q){
		(*q)->StopTask();
	}
	//queue could still exist somewhere else, but inactive. their input/output
	//are frozen in their last state
}

bool Manager::RegisterOutput(const std::string& name , Output::Ptr out){
	return d_outputs.Register(name,out);
}

bool Manager::RegisterInput(const std::string& name , Input::Ptr in){
	return d_inputs.Register(name,in);
}

void Manager::InitializeQueues(){
	d_ioEvent = NativeHolder<RT_EVENT>(new RT_EVENT());
	d_idleQueues.reserve(d_queues.size());
	int res = rt_event_create(d_ioEvent.get(),IOEventName,0,EV_PRIO);
	if(res == -EEXIST ){
		throw std::runtime_error("Synchronization signal ressource is already "
				"registered in the system. Only one instance of Manager could "
				"be run on the whole system.");
	}
	if( res != 0 ){
		throw rx::NativeError("Error during synchro event initialization",-res);
	}
	d_logEvent = NativeHolder<RT_EVENT>(new RT_EVENT());
	res = rt_event_create(d_logEvent.get(),LogEventName,0,EV_FIFO);
	if(res){
		throw rx::NativeError("Error during log event creation",-res);
	}
	if(d_queues.size() == MaxNumberOfQueue){
		d_allEventMask = 0;
		--d_allEventMask;
	} else {
		d_allEventMask = (1UL << (d_queues.size())) - 1;
	}
	d_lastEventMask = 0;

	d_inBuffer.reserve(d_inputs.Size());
	d_outBuffer.reserve(d_outputs.Size());

	for(QueueById::iterator q = d_queues.begin();
	    q != d_queues.end();
	    ++q){
		(*q)->StartTask();
	}

	for(LogQueueById::iterator lq = d_logQueues.begin();
	    lq != d_logQueues.end();
	    ++lq){
		(*lq)->SetName(d_inputs.IONames(),d_outputs.IONames());
		(*lq)->Init();
		(*lq)->StartTask();
	}
}


void Manager::StartIdleQueues(){
	//All bit at 0 have finsihed and thus ready to be started, bit to 1 should
	//not be set
	rt_event_signal(d_ioEvent.get(),d_lastEventMask ^ d_allEventMask);
}


const Manager::ListOfQueue & Manager::IdleQueues(){
	if(!d_ioEvent){
		throw std::logic_error("You did not Manager::InitializeQueues() !!!");
	}
	rt_event_wait(d_ioEvent.get(),0,&d_lastEventMask,EV_ANY,TM_NONBLOCK);
	d_idleQueues.clear();
	if(d_lastEventMask == d_allEventMask){
		return d_idleQueues;
	}

	for(unsigned long i = 0; i < d_queues.size(); ++i){
		if((d_lastEventMask & (1UL << i)) == 0){
			//queue i have finished
				d_idleQueues.push_back(d_queues[i]);
		}
	}
	return d_idleQueues;
}

rtio::Output::Ptr Manager::GetOutput(const std::string & name){
	return d_outputs.IO(name);
}

rtio::Input::Ptr  Manager::GetInput (const std::string & name){
	return d_inputs.IO(name);
}

const Queue::OutputByName & Manager::Outputs() const{
	return d_outputs.IOByName();
}

const Queue::InputByName & Manager::Inputs() const{
	return d_inputs.IOByName();
}

void Manager::BufferData(double time){
	d_inBuffer = d_inputs.IOData();
	d_outBuffer = d_outputs.IOData();
	d_timeBuffer = time;
}


void Manager::StartIdleLogQueues(){
	unsigned long globalMask;
	if(d_logQueues.size() < MaxNumberOfQueue){
		 globalMask = (1UL << d_logQueues.size()) - 1;
	} else {
		globalMask = 0;
		--globalMask;
	}

	unsigned long idleQueueMask;
	rt_event_wait(d_logEvent.get(),0,&idleQueueMask,EV_ANY,TM_NONBLOCK);

	for(unsigned int i = 0; i < d_logQueues.size(); ++i){
		if(((1UL << i) & idleQueueMask) == 0){
			//logQueue is idle;
			d_logQueues[i]->SetTime(d_timeBuffer);
			d_logQueues[i]->SetData(d_inBuffer,d_outBuffer);
		}
	}

	//start all idle queues again  ....
	rt_event_signal(d_logEvent.get(),globalMask & (~idleQueueMask));

}
