/**
 * \file RTIOManager.h
 *
 *  \date Dec 4, 2012
 *  \author Alexandre Tuleu
 */

#ifndef ROBO_XENO_IO_MANAGER_H_
#define ROBO_XENO_IO_MANAGER_H_

#include <vector>

#include "Queue.h"
#include "LogQueue.h"
#include "IOHolder.h"

#include <native/event.h>
#include <stdexcept>
#include <sstream>

#include <robo-xeno/NativeHolder.h>

namespace rtio{
class Manager {
	DISABLE_COPY(Manager)
	DISABLE_ASSIGNEMENT(Manager)
public:
	typedef std::vector<Queue::Ptr>  ListOfQueue;


	Manager();
	virtual ~Manager();

	void InitializeQueues();

	template <class QueueType>
	Queue::Ptr AddQueue(int priority = 0, Queue::Mode m = Queue::NonRealtime);

	template <class LogQueueType>
	LogQueue::Ptr AddLogQueue();

	const ListOfQueue & IdleQueues();
	void StartIdleQueues();

	Output::Ptr GetOutput(const std::string & name);
	Input::Ptr  GetInput (const std::string & name);

	const Queue::OutputByName & Outputs() const;
	const Queue::InputByName & Inputs() const;

	void BufferData(double time);
	void StartIdleLogQueues();


	class QueueAccessor {
	private :
		static bool RegisterOutput(Manager & io,const std::string & name, Output::Ptr out);
		static bool RegisterInput(Manager & io, const std::string & name, Input::Ptr out);
		friend class rtio::Queue;
	};
	friend class QueueAccessor;

	const static char * IOEventName;
	const static char * LogEventName;
	const static Queue::ID MaxNumberOfQueue;
private :
	typedef std::tr1::shared_ptr<Queue>    QueuePtr;
	typedef std::tr1::shared_ptr<LogQueue> LogQueuePtr;
	typedef std::vector<QueuePtr>          QueueById;
	typedef std::vector<LogQueuePtr>       LogQueueById;
	typedef unsigned long                  EventMask;






	bool RegisterOutput(const std::string & name, Output::Ptr out);
	bool RegisterInput(const std::string & name, Input::Ptr in);






	IOHolder<rtio::Output> d_outputs;
	IOHolder<rtio::Input>  d_inputs;
	ListOfQueue         d_idleQueues;
	QueueById           d_queues;
	LogQueueById        d_logQueues;
	NativeHolder<RT_EVENT> d_ioEvent;
	NativeHolder<RT_EVENT> d_logEvent;
	EventMask           d_allEventMask;
	EventMask           d_lastEventMask;
	std::vector<double> d_inBuffer,d_outBuffer;
	double              d_timeBuffer;
};




inline bool Manager::QueueAccessor::RegisterOutput(Manager & io,
                                                   const std::string & name,
                                                   Output::Ptr out){
	return io.RegisterOutput(name,out);
}

inline bool Manager::QueueAccessor::RegisterInput(Manager & io,
                                                  const std::string & name,
                                                  Input::Ptr in){
	return io.RegisterInput(name,in);
}
template <class QueueType>
inline Queue::Ptr Manager::AddQueue(int priority, Queue::Mode m){
	if(d_ioEvent){
		throw std::logic_error("You can only rtio::Manager::AddQueue<T> *before* rtio::Manager::InitializeQueues()");
	}

	if(d_queues.size() >= MaxNumberOfQueue){
		std::ostringstream os;
		os << "There is maximal number of " << MaxNumberOfQueue << " RTIO::Queue"
		   << " accepted on this system. You cannot register another one.";
		throw std::overflow_error(os.str());
	}



	Queue::ID newId(d_queues.size());
	QueuePtr q(new QueueType(*this,newId,priority,m));
	//d_queueIDs.insert(std::make_pair(q,newId));
	d_queues.push_back(q);
	return q;
}


template <class LogQueueType>
inline LogQueue::Ptr Manager::AddLogQueue(){
	if(d_logEvent){
		throw std::logic_error("You can only rtio::Manager::AddLogQueue<T> *before* rtio::Manager::InitializeQueues()");
	}

	if(d_logQueues.size() >= MaxNumberOfQueue){
		std::ostringstream os;
		os << "maximum number of allowed log queues (" << MaxNumberOfQueue << " reached.";
		throw std::overflow_error(os.str());
	}

	LogQueue::ID newId(d_logQueues.size());
	LogQueue::Ptr lq(new LogQueueType(newId));
	d_logQueues.push_back(lq);
	return lq;
}

} //namespace rt
#endif // ROBO_XENO_RTIOMANAGER_H_
