/**
 * \file ManagerUTest.cpp
 *
 *  \date Dec 4, 2012
 *  \author Alexandre Tuleu
 */

#include "ManagerUTest.h"

#include <stdexcept>
#include <unistd.h>

namespace rtio{

void ManagerUTest::SetUp(){
}

void ManagerUTest::TearDown(){

}


void ManagerSyncUTest::SetUp() {


	for(unsigned int i =0 ; i < 2; ++i){
		Queue::Ptr q = d_manager.AddQueue<MockQueue>();
		d_q[i] = std::tr1::static_pointer_cast<MockQueue,Queue>(q);

		d_in[2*i]     = MockInput::Ptr(new MockInput());
		d_in[2*i + 1] = MockInput::Ptr(new MockInput());

		d_out[2*i]     = MockOutput::Ptr(new MockOutput());
		d_out[2*i + 1] = MockOutput::Ptr(new MockOutput());
		std::string number("1");
		number.at(0) += (char)i;
		d_q[i]->RegisterInput("a" + number,
		                      std::tr1::static_pointer_cast<Input,MockInput>(d_in[2 * i]));
		d_q[i]->RegisterInput("b" + number,
		                      std::tr1::static_pointer_cast<Input,MockInput>(d_in[2 * i + 1]));

		d_q[i]->RegisterOutput("a" + number,
		                       std::tr1::static_pointer_cast<Output,MockOutput>(d_out[2 * i]));
		d_q[i]->RegisterOutput("b" + number,
		                       std::tr1::static_pointer_cast<Output,MockOutput>(d_out[2 * i + 1]));


	}

}

void ManagerSyncUTest::TearDown() {

}

TEST_F(ManagerUTest,CouldNotAddQueueAfterInitialization){
	d_manager.InitializeQueues();
	EXPECT_THROW({
		d_manager.AddQueue<MockQueue>();
	}, std::logic_error);
}

TEST_F(ManagerUTest,HasAMaximalNumberOFQueue){
	EXPECT_NO_THROW({
		for(unsigned int i = 0 ; i < Manager::MaxNumberOfQueue ; ++i){
			d_manager.AddQueue<MockQueue>();
		}
	});

	EXPECT_THROW({
		d_manager.AddQueue<MockQueue>();
	}, std::overflow_error );
}




TEST_F(ManagerSyncUTest,SynchronizationIsOK){
	//In this scenario, we use two queue. the first will be slow,
	//and the second super fast.
	//We use semaphore to control our test state. because we have several
	// thread !!!


	//first queue will be blocked during first sync
	EXPECT_CALL(*d_q[0], PreSend());
	//we wait for our semaphore to be unlocked ....
	EXPECT_CALL(*d_q[0], PreRead()).WillOnce(WaitSemaphore(d_q[0]->MySemaphore()));
	//and we increment the semaphore ofq all queues
	EXPECT_CALL(*d_q[0],PostRead()).WillOnce(PostSemaphore(MockQueue::Semaphore()));
	EXPECT_CALL(*d_in[0], Read());
	EXPECT_CALL(*d_in[1], Read());
	EXPECT_CALL(*d_out[0],Send());
	EXPECT_CALL(*d_out[1],Send());

	//second queue will never be blocked
	EXPECT_CALL(*d_q[1], PreSend()).Times(2);
	EXPECT_CALL(*d_q[1], PreRead()).Times(2);
	EXPECT_CALL(*d_q[1],PostRead()).Times(2).WillRepeatedly(PostSemaphore(MockQueue::Semaphore()));
	EXPECT_CALL(*d_in[2] ,Read()).Times(2);
	EXPECT_CALL(*d_in[3] ,Read()).Times(2);
	EXPECT_CALL(*d_out[2],Send()).Times(2);
	EXPECT_CALL(*d_out[3],Send()).Times(2);

	//in xenomai, semaphore will be locked only if we are shadowing (no effect on emulation)
	RT_TASK this_task;
	rt_task_shadow(&this_task,NULL,0,T_FPU);

	d_manager.InitializeQueues();

	//starts the two queue
	d_manager.StartIdleQueues();
	RT_SEM_INFO info;
	//only second queue will finish, we wait for that moment
	rt_sem_p(MockQueue::Semaphore(),1e8);
	rt_sem_inquire(MockQueue::Semaphore(),&info);
	//here only one queue should have finished
	EXPECT_EQ(0,info.count);

	//we consume data
	const Manager::ListOfQueue & queue = d_manager.IdleQueues();
	//only queue 2 should have been posted ready
	EXPECT_EQ(1,queue.size());
	EXPECT_EQ(d_q[1],queue[0]);
	//we release the first queue, before syncing again
	rt_sem_v(d_q[0]->MySemaphore());
	//we wait for the first queue to finish
	rt_sem_p(MockQueue::Semaphore(),1e8);
	//now we sync again, since first queue did not finish before InputsReady(),
	//it should not be started again
	d_manager.StartIdleQueues();



	rt_sem_p(MockQueue::Semaphore(),1e8);
	rt_sem_inquire(MockQueue::Semaphore(),&info);
	//no more than one queue should have been processed
	EXPECT_EQ(0,info.count);

	const Manager::ListOfQueue & queue2 = d_manager.IdleQueues();
	ASSERT_EQ(2,queue2.size());

}

TEST_F(ManagerSyncUTest,CanAccessIOGloablly){
	ASSERT_EQ(4,d_manager.Inputs().size());
	ASSERT_EQ(4,d_manager.Outputs().size());

	EXPECT_NE(rtio::Output::Ptr(),d_manager.GetOutput("a1"));
	EXPECT_NE(rtio::Output::Ptr(),d_manager.GetOutput("a2"));
	EXPECT_NE(rtio::Output::Ptr(),d_manager.GetOutput("b1"));
	EXPECT_NE(rtio::Output::Ptr(),d_manager.GetOutput("b2"));

	EXPECT_NE(rtio::Input::Ptr(),d_manager.GetInput("a1"));
	EXPECT_NE(rtio::Input::Ptr(),d_manager.GetInput("a2"));
	EXPECT_NE(rtio::Input::Ptr(),d_manager.GetInput("b1"));
	EXPECT_NE(rtio::Input::Ptr(),d_manager.GetInput("b2"));
}
}
