/**
 * \file RTQueue.cpp
 *
 *  \date Dec 4, 2012
 *  \author Alexandre Tuleu
 */

#include "Queue.h"
#include "Manager.h"
#include <robo-xeno/NativeError.h>
#include <stdexcept>
#include <cstring>
using namespace rtio;

Queue::Queue(rtio::Manager & manager , Queue::ID id,int priority , Mode m)
		: d_manager(manager)
		, d_task(new RT_TASK())
		, d_priority(priority)
		, d_mode(m)
		, d_id(id)
		, d_inputs()
		, d_outputs(){
		int res;
		if((res = rt_task_create(d_task.get(),NULL,0,d_priority,0))){
			throw rx::NativeError("Could not create task",-res);
		}

	}

Queue::~Queue(){
}


Manager & Queue::GetManager() {
	return d_manager;
}

void Queue::RegisterInput(const std::string & name,Input::Ptr input){
	if(!rtio::Manager::QueueAccessor::RegisterInput(d_manager,name,input)){
		throw std::runtime_error("Could not register input : '" + name + "' is already registered");
	}
	d_inputs.insert(std::make_pair(name,input));
}
void Queue::RegisterOutput(const std::string & name,Output::Ptr output){
	if(!rtio::Manager::QueueAccessor::RegisterOutput(d_manager,name,output)){
		throw std::runtime_error("Could not register output : '" + name + "' is already registered");
	}

	d_outputs.insert(std::make_pair(name,output));
}


const Queue::OutputByName & Queue::Outputs() const{
	return d_outputs;
}

const Queue::InputByName & Queue::Inputs() const{
	return d_inputs;
}


void Queue::PreSend(){
	//do nothing
}
void Queue::PreRead(){
	//do nothing
}
void Queue::PostRead(){
	//do nothing
}

void Queue::StartTask(){
	if(!d_task){
		//here the task is already stopped destroy, so we do not start it !
		return;
	}
	//reinterpret_cast is ugly, but the standard insure that we will get the
	//same object after the second one.
	int res = rt_task_start(d_task.get(),&Queue::RtBody,reinterpret_cast<void*>(this));
	if(res){
		throw rx::NativeError("Could not start task" ,-res);
	}
}


void Queue::StopTask(){
	//the shared ptr is released, and will call rt_task_delete ...
	d_task = NativeHolder<RT_TASK>();
}

void Queue::RtBody(void * itself){
	//authorized because we are private and we could make sure that it was
	//firstly reinterpreted_casted
	reinterpret_cast<Queue*>(itself)->Loop();
}


void Queue::Loop(){
	RT_EVENT event;
	rt_event_bind(&event,Manager::IOEventName,TM_INFINITE);

	unsigned long myBit = (1UL << d_id);
	unsigned long mask;
	while(true){
		int res = rt_event_wait(&event,myBit,&mask,EV_ANY,TM_INFINITE);
		if(res != 0){
			break;
		}

		if(d_mode == NonRealtime){
			rt_task_set_mode(T_PRIMARY,0,NULL);
		}

		PreSend();

		for(OutputByName::iterator o = d_outputs.begin();
		    o != d_outputs.end();
		    ++o){
			o->second->Send();
		}


		PreRead();

		for(InputByName::iterator i = d_inputs.begin();
		    i != d_inputs.end();
		    ++i){
			i->second->Read();
		}

		PostRead();

		rt_event_clear(&event,myBit,0);

	}
}


