/**
 * \file event.c
 *
 *  \date Dec 4, 2012
 *  \author Alexandre Tuleu
 */


#include "event.h"

#include <map>
#include <string>
#include <cstdlib>
#include <cstring>

pthread_mutex_t EventMutex = PTHREAD_MUTEX_INITIALIZER;
typedef std::multimap<std::string,RT_EVENT * > EventByName;
EventByName named_events;

robo_xeno_rt_event::robo_xeno_rt_event() :  name(0) , mask(0) {};


extern "C"{


int rt_event_create (RT_EVENT *event, const char *name, unsigned long ivalue, int mode){
	if(name != NULL){

		pthread_mutex_lock(&EventMutex);
		bool exists(named_events.find(name) != named_events.end());
		pthread_mutex_unlock(&EventMutex);

		if(exists){
			return -1;
		}
	}

	event->mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
	pthread_mutex_init(event->mutex,NULL);
	event->mask = (unsigned long *)malloc(sizeof(unsigned long));
	event->name = NULL;

	if(name){
		event->name = (char *) malloc(strlen(name) + 1);
		strcpy(event->name,name);

		pthread_mutex_lock(&EventMutex);
		named_events.insert(std::make_pair(name,event));
		pthread_mutex_unlock(&EventMutex);
	}


	*(event->mask) = ivalue;
	return 0;
}

int rt_event_delete (RT_EVENT *event){
	pthread_mutex_lock(&EventMutex);
	if(event->mutex){
		pthread_mutex_destroy(event->mutex);
		free(event->mutex);
	}
	if(event->mask){
		free(event->mask);
	}
	if(event->name){

		std::pair<EventByName::iterator,EventByName::iterator> range =
		named_events.equal_range(event->name);
		free(event->name);
		for(EventByName::iterator e = range.first;
		    e != range.second;
		    ++e){
			e->second->mutex = 0;
			e->second->mask  = 0;
			e->second->name  = 0;
		}
		named_events.erase(range.first,range.second);
	}

	pthread_mutex_unlock(&EventMutex);
	return 0;
}



int rt_event_wait (RT_EVENT *event, unsigned long mask,
                   unsigned long *mask_r, int mode, RTIME timeout){

	unsigned long actual_mask;

	if(timeout == TM_NONBLOCK){
		pthread_mutex_lock(&EventMutex);
		if(!event->mutex){
			pthread_mutex_unlock(&EventMutex);
			return -1;
		}
		pthread_mutex_lock(event->mutex);
		actual_mask = *(event->mask);
		pthread_mutex_unlock(event->mutex);
		pthread_mutex_unlock(&EventMutex);
		bool result = mode == EV_ANY ? actual_mask & mask
		                             : (actual_mask & mask) == mask;
		//this will segfault, but like xenomai 2.5.6
		*mask_r = actual_mask;

		if(result){
			return 0;
		} else {
			return -1;
		}
	}

	while(true){
		pthread_mutex_lock(&EventMutex);
		if(!event->mutex){
			pthread_mutex_unlock(&EventMutex);
			return -1;
		}
		pthread_mutex_lock(event->mutex);
		actual_mask = *(event->mask);
		pthread_mutex_unlock(event->mutex);
		pthread_mutex_unlock(&EventMutex);
		if(mode == EV_ANY && actual_mask & mask){
			break;
		} else if((actual_mask & mask) == mask) {
			break;
		}
		usleep(10000);
	}

	//this will segfault, but like xenomai 2.5.6
	*mask_r = actual_mask;

	return 0;
}

int rt_event_clear (RT_EVENT *event, unsigned long mask, unsigned long *mask_r){
	pthread_mutex_lock(&EventMutex);
	if(!event->mutex){
		pthread_mutex_unlock(&EventMutex);
		return -1;
	}
	pthread_mutex_lock(event->mutex);
	if(mask_r){
		*mask_r = *(event->mask);
	}
	*(event->mask) = *(event->mask) & ~mask;
	pthread_mutex_unlock(event->mutex);
	pthread_mutex_unlock(&EventMutex);
	return 0;
}

int rt_event_bind (RT_EVENT *event, const char *name, RTIME timeout){
	if(name == NULL){
		return -2;
	}

	while(true){
		pthread_mutex_lock(&EventMutex);
		EventByName::iterator e = named_events.find(name);
		if(e != named_events.end()){


			event->name = e->second->name;
			event->mutex = e->second->mutex;
			event->mask = e->second->mask;

			named_events.insert(std::make_pair(name,event));
			pthread_mutex_unlock(&EventMutex);
			return 0;
		}
		pthread_mutex_unlock(&EventMutex);
		usleep(10000);
	}
	return 0;
}

int rt_event_signal (RT_EVENT *event, unsigned long mask){
	pthread_mutex_lock(&EventMutex);
	if(!event->mutex){
		pthread_mutex_unlock(&EventMutex);
		return -1;
	}
	pthread_mutex_lock(event->mutex);
	*(event->mask) = *(event->mask) | mask;
	pthread_mutex_unlock(event->mutex);
	pthread_mutex_unlock(&EventMutex);
	return 0;
}

}
